// $Id: bdos.c,v 1.10 2011-04-08 23:11:31 drmiller Exp $

#include "bdos.h"
#include "z80_sys.h"
#include "z80_exec.h"

#include "bdos_impl.h"

static bdos_params_t _curparm;

void bdos_init(z80_sys_t *sys) {
	bdos_impl_init(sys);

	bdos_func_t *f = f = bdos_funcs[13];	// reset BDOS
	if (f) f(sys, NULL);	// we hope this ignores 'parm'...
}

static int bdos(z80_sys_t *sys) {
	_curparm.de = *(get_rp(sys, rp_DE));
	_curparm.bc = *(get_rp(sys, rp_BC));
	_curparm.a = 0;
	_curparm.hl = 0;
	bdos_func_t *f;
	f = bdos_funcs[_curparm.bc & 0x00ff];
	if (!f) {
		sys->fault(sys, "Invalid BDOS function code");
	}
	int rc = f(sys, &_curparm);
	if (sys->run == 0) return 1;
	if (rc < 0) {
		sys->fault(sys, "BDOS function fatal");
	}
#ifdef TRACE
	if (sys->trace) {
		fprintf(sys->trc_fp, "BDOS function %d DE=%04x / A=%02x HL=%04x\n",
				_curparm.bc & 0x00ff, _curparm.de, _curparm.a, _curparm.hl);
	}
#endif // TRACE
	*(get_rp(sys, rp_HL)) = _curparm.hl;
	*(get_gp_A(sys)) = _curparm.hl & 0x00ff;
	return (sys->cpu.pc == 0x0000);
}

static int bios(z80_sys_t *sys) {
	int func = (sys->cpu.pc & 0x00ff) / 3;
	_curparm.de = *(get_rp(sys, rp_DE));
	_curparm.bc = *(get_rp(sys, rp_BC));
	_curparm.a = 0;
	_curparm.hl = 0;
	bios_func_t *f;
	f = bios_funcs[func];
	if (!f) {
		sys->fault(sys, "Invalid BIOS vector called");
	}
	int rc = f(sys, &_curparm);
	if (sys->run == 0) return 1;
	if (rc < 0) {
		sys->fault(sys, "BIOS function fatal");
	}
#ifdef TRACE
	if (sys->trace) {
		fprintf(sys->trc_fp, "BIOS function %d A=%02x BC=%04x DE=%04x HL=%04x\n",
			func, _curparm.a, _curparm.bc, _curparm.de, _curparm.hl);
	}
#endif // TRACE
	*(get_rp(sys, rp_HL)) = _curparm.hl;	// cannot change?
	*(get_gp_A(sys)) = _curparm.a;
	return (sys->cpu.pc == 0x0000);
}

// return 0 if ok to execute at PC
int bdos_trap(z80_sys_t *sys) {
	if (sys->cpu.pc == BDOS_ENTRY) {
		return bdos(sys);
	} else if ((sys->cpu.pc >> 8) == (BIOS_START >> 8)) {
		return bios(sys);
	}
	return 0;
}
