// $Id: bdos_impl.c,v 1.16 2011-04-08 22:45:51 drmiller Exp $

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <dirent.h>
#include <fnmatch.h>
#include <errno.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <signal.h>

#include "bdos.h"
#include "bdos_impl.h"
#include "z80_sys.h"
#include "z80_instr.h" // just need ret_op()...
#include "z80_exec.h"

void bdos_intr(int sig) {
	// nothing to do, just wanted to interrupt syscall...
}

int intr_read(z80_sys_t *sys, int fd, void *buf, size_t len) {
	struct sigaction old;
	struct sigaction new;

	new.sa_handler = bdos_intr;
	new.sa_flags = 0;
	sigemptyset(&new.sa_mask);
	sigaction(SIGINT, &new, &old);
	int rc = read(fd, buf, len);
	int intr = (rc < 0 && errno == EINTR);
	sigaction(SIGINT, &old, NULL);
	if (intr) {
		sys->run = 0;
	}
	return rc;
}

char *intr_fgets(z80_sys_t *sys, char *s, int size, FILE *stream) {
	struct sigaction old;
	struct sigaction new;

	new.sa_handler = bdos_intr;
	new.sa_flags = 0;
	sigemptyset(&new.sa_mask);
	sigaction(SIGINT, &new, &old);
	char *t = fgets(s, size, stream);
	int intr = (!t && errno == EINTR);
	sigaction(SIGINT, &old, NULL);
	if (intr) {
		sys->run = 0;
	}
	return t;
}

// --------------- BIOS implementation ---------------------

static uint8_t _conchr = 0;

static int bios_warm(z80_sys_t *sys, bdos_params_t *parm) {
	sys->cpu.pc = 0x0000;	// this won't work!
	return 0;
}
static int bios_const(z80_sys_t *sys, bdos_params_t *parm) {
	int st = 0;
//	if (_conchr) {
//		parm->a = 0xff;
//		return 0;
//	}
	int rc = ioctl(0, FIONREAD, &st);
	if (st) {
		parm->a = 0xff;
		return 0;
	}
	parm->a = 0;
	return 0;
}
static int bios_conin(z80_sys_t *sys, bdos_params_t *parm) {
	intr_read(sys, 0, &parm->a, 1);
	return 0;
}
static int bios_conout(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t c = parm->bc & 0x00ff;
	write(1, &c, 1);
	return 0;
}
static int bios_list(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t c = parm->bc & 0x00ff;
	// print(1, &c, 1);
	return 0;
}
static int bios_listst(z80_sys_t *sys, bdos_params_t *parm) {
	parm->a = 0xff;	// ready
	return 0;
}

bios_func_t *bios_funcs[255/3] = {
// [0] = bios_cold,	// should never be called
[1] = bios_warm,
[2] = bios_const,
[3] = bios_conin,
[4] = bios_conout,
[5] = bios_list,
// [6] = bios_auxout,
// [7] = bios_auxin,
// [8] = bios_home,
// [9] = bios_seldsk,
// [10] = bios_settrk,
// [11] = bios_setsec,
// [12] = bios_setdma,
// [13] = bios_read,
// [14] = bios_write,
[15] = bios_listst,
// [16] = bios_sectrn,
// [17] = bios_conost,
// [18] = bios_auxist,
// [19] = bios_auxost,
};

// --------------- BDOS implementation ---------------------

// this should be configurable, commandline possibly
#define CPM_PREFIX	"./cpm_disks"

#define OFFSET_OF(s,e)	((uintptr_t)&((s *)0)->e)

typedef struct {
	uint8_t drv;
	uint8_t name[11];
	uint8_t ex;
	uint8_t s1[2];
	uint8_t rc;
	union {
		uint8_t d0[16];
		int fd;
	};
	uint8_t cr;	// current record
	uint8_t rr[3];	// random record num
} __attribute__((packed)) fcb_impl_t;

typedef struct {
	DIR *dir;
	char pat[32];
	uint8_t drv;
	uint8_t usr;
	uint8_t ext;
	off_t size;
	char dirpath[256];
} search_t;

typedef struct {
	uint16_t spt;
	uint8_t bsh;
	uint8_t blm;
	uint8_t exm;
	uint16_t dsm;
	uint16_t drm;
	uint8_t al0;
	uint8_t al1;
	uint16_t cks;
	uint16_t off;
} __attribute__((packed)) bdos_dpb_t;

#define DEF_BLS_SH	14	// 2^14 = 16K
#define DEF_BLS		(1 << DEF_BLS_SH)
#define DEF_NBLOCKS	128	// keep alloc vec small, disk size 2M

typedef struct {
	uint8_t jmp[3];
	bdos_dpb_t dpb;
	uint8_t alloc[DEF_NBLOCKS / 8];
} bdos_page_t;

static uint16_t _curdma;
static uint8_t *_curdma_real;
static uint16_t _curdsk;
static uint16_t _curusr;
static uint16_t _curchr;
static search_t _cur_search = {NULL};

static uint32_t _cur_ro = 0;
static uint32_t _cur_log = 0;

static bdos_dpb_t _cur_dpb = {
spt:	36,	// anything
bsh:	DEF_BLS_SH - 7,
blm:	(DEF_BLS / 128) - 1,
exm:	16 / (16384 / DEF_BLS) - 1,
dsm:	DEF_NBLOCKS-1,
drm:	0,	// anything
al0:	0,	// anything
al1:	0,	// anything
cks:	0,	// anything
off:	0,	// anything
};

static void _get_fn(char *dst, fcb_impl_t *fcb) {
	int x;
	for (x = 0; x < 8 && fcb->name[x] != ' '; ++x) {
		if (fcb->name[x] >= 'A') *dst++ = fcb->name[x] | 0x20;
		else *dst++ = fcb->name[x];
	}
	for (x = 8; x < 11 && fcb->name[x] != ' '; ++x) {
		if (x == 8) *dst++ = '.';
		if (fcb->name[x] >= 'A') *dst++ = fcb->name[x] | 0x20;
		else *dst++ = fcb->name[x];
	}
	*dst++ = '\0';
}

static void _get_afn(char *dst, fcb_impl_t *fcb, uint8_t usr) {
	int x;
	if (usr) {
		dst += sprintf(dst, "%d:", usr);
	}
	for (x = 0; x < 8 && fcb->name[x] != ' '; ++x) {
		if (fcb->name[x] == '?') {
			*dst++ = '*';
			break;
		}
		if (fcb->name[x] >= 'A') *dst++ = fcb->name[x] | 0x20;
		else *dst++ = fcb->name[x];
	}
	for (x = 8; x < 11 && fcb->name[x] != ' '; ++x) {
		if (x == 8) *dst++ = '.';
		if (fcb->name[x] == '?') {
			*dst++ = '*';
			break;
		}
		if (fcb->name[x] >= 'A') *dst++ = fcb->name[x] | 0x20;
		else *dst++ = fcb->name[x];
	}
	*dst++ = '\0';
}

static void _copyout_dir(uint8_t *dma, char *name) {
	fcb_impl_t *fcb = (fcb_impl_t *)dma;
	char *t;
	t = strchr(name, ':');
	uint8_t u;
	if (t) {
		u = atoi(name);
		++t;
	} else {
		u = 0;
		t = name;
	}
	fcb->drv = u;	// user code , not drive
	int x = 0;
	while (*t && *t != '.' && x < 8) {
		fcb->name[x++] = toupper(*t++);
	}
	while (x < 8) fcb->name[x++] = ' ';
	while (*t && *t != '.') ++t;
	if (*t == '.') ++t;
	while (*t && *t != '.' && x < 11) {
		fcb->name[x++] = toupper(*t++);
	}
	while (x < 11) fcb->name[x++] = ' ';
	fcb->ex = 0;
	fcb->s1[0] = 0;
	fcb->s1[1] = 0;
	fcb->rc = 0;
	memset(fcb->d0, 0, sizeof(fcb->d0));
}

static int _start_search(fcb_impl_t *fcb, search_t *search) {
	uint8_t u = _curusr;
	uint8_t d = fcb->drv;
	if (d == '?') {
		d = _curdsk;
		search->pat[0] = '*';
		search->pat[1] = '\0';
	} else {
		if (!d) d = _curdsk;
		else --d;
		_get_afn(search->pat, fcb, u);
	}
	(void)sprintf(search->dirpath, "%s/%c:", CPM_PREFIX, d + 'A');
	search->drv = d;
	search->usr = u;
	search->ext = fcb->ex;
	search->dir = opendir(search->dirpath);
	if (!search->dir) {
		return -1;
	}
	return 0;
}

static char *_do_search(search_t *search) {
	struct dirent *dir;
	int found = 0;
	do {
		dir = readdir(search->dir);
		if (!dir) {
			return NULL;
		}
		if (dir->d_name[0] == '.') continue;
		//if (dir->d_type != DT_REG) continue;
		if (fnmatch(search->pat, dir->d_name, 0) == 0) break;
	} while (!found);
	if (search->ext == '?') {
		// return size of file... by some definition...
		struct stat stb;
		char buf[1024];
		sprintf(buf, "%s/%s", search->dirpath, dir->d_name);
		stat(buf, &stb);
		search->size = stb.st_size;
	}
	return dir->d_name;
}

static void _seek_file(fcb_impl_t *fcb) {
	off_t r = fcb->rr[0] | (fcb->rr[0] << 8) | (fcb->rr[0] << 16);
	r *= 128;
	(void)lseek(fcb->fd, r, SEEK_SET);
	fcb->s1[0] = fcb->cr;
}

static int _open_file(fcb_impl_t *fcb) {
	char buf[32];
	char path[256];
	uint8_t d;
	_get_fn(buf, fcb);
	d = fcb->drv;
	if (!d) {
		d = _curdsk;
	} else {
		--d;
	}
	uint8_t u = _curusr;
	if (u) {
		sprintf(path, "%s/%c:/%d:%s", CPM_PREFIX, d + 'A', u, buf);
	} else {
		sprintf(path, "%s/%c:/%s", CPM_PREFIX, d + 'A', buf);
	}
	int fd = open(path, O_RDWR);
	if (fd < 0 && u) {
		sprintf(path, "%s/%c:/%s", CPM_PREFIX, d + 'A', buf);
		fd = open(path, O_RDWR);
	}
	if (fd < 0) {
		// don't know exactly why... _cur_log &= ~(1 << d);
		return -1;
	}
	_cur_log |= (1 << d); // drive must be valid
	fcb->ex = 0;
	fcb->rc = 0;
	fcb->fd = fd;
	struct stat stb;
	fstat(fd, &stb);
	off_t len = stb.st_size;
	len = (len + 127) / 128; // num records
	fcb->rc = len & 0x7f;
	fcb->s1[0] = 0;
	fcb->s1[1] = 0x80; // flag if close needs to update...
	return 0;
}

static int _bdos_inval(z80_sys_t *sys, bdos_params_t *parm) {
	fprintf(stderr, "Invalid BDOS func %d DE=%04x\n", parm->bc & 0x00ff, parm->de);
	return -1;
}

static int _bdos_unimpl(z80_sys_t *sys, bdos_params_t *parm) {
	fprintf(stderr, "Unimplemented BDOS func %d DE=%04x\n", parm->bc & 0x00ff, parm->de);
	return 0;
}

static int bdos_getdpb(z80_sys_t *sys, bdos_params_t *parm) {
	parm->hl = BDOS_START + OFFSET_OF(bdos_page_t,dpb);
	return 0;
}
static int bdos_prot(z80_sys_t *sys, bdos_params_t *parm) {
	_cur_ro |= (1 << _curdsk);
	return 0;
}
static int bdos_resdrv(z80_sys_t *sys, bdos_params_t *parm) {
	_cur_ro &= ~(1 << _curdsk);
	return 0;
}
static int bdos_getro(z80_sys_t *sys, bdos_params_t *parm) {
	parm->hl = _cur_ro;
	return 0;
}
static int bdos_getall(z80_sys_t *sys, bdos_params_t *parm) {
	parm->hl = BDOS_START + OFFSET_OF(bdos_page_t,alloc);
	return 0;
}
static int bdos_getlog(z80_sys_t *sys, bdos_params_t *parm) {
	parm->hl = _cur_log;
	return 0;
}

static int bdos_reset(z80_sys_t *sys, bdos_params_t *parm) {
	sys->cpu.pc = 0x0000;
	return 0;
}

static int bdos_conin(z80_sys_t *sys, bdos_params_t *parm) {
	int rc = bios_conin(sys, parm);
	if (rc) return rc;
	parm->hl = parm->a;
	return 0;
}

static int bdos_conout(z80_sys_t *sys, bdos_params_t *parm) {
	parm->bc = parm->de;
	int rc = bios_conout(sys, parm);
	return rc;
}

static int bdos_getiob(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, 0x0003);
	parm->hl = *m;
	return 0;
}

static int bdos_setiob(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, 0x0003);
	uint8_t e = parm->de & 0x00ff;
	*m = e;
	return 0;
}

static int bdos_msgout(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *s, *m;
	m = mem_get_addr(&sys->mem, parm->de);
	s = strchr(m, '$');
	if (!s) {
		fprintf(stderr, "Un-terminated string in BDOS function 9\n");
	} else {
		write(1, m, s - m);
	}
	return 0;
}

static int bdos_setdma(z80_sys_t *sys, bdos_params_t *parm) {
	_curdma = parm->de;
	_curdma_real = mem_get_addr(&sys->mem, parm->de);
	return 0;
}

static int bdos_seldsk(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t d = parm->de & 0x00ff;
	if (_curdsk == d) {
		return 0;
	}
	_curdsk = d;
	char buf[256];
	struct stat stb;
	sprintf(buf, "%s/%c:", CPM_PREFIX, d + 'A');
	int rc = stat(buf, &stb);
	if (rc < 0 || !S_ISDIR(stb.st_mode)) {
		_cur_log &= ~(1 << d);
		parm->hl = 0x0004;
		return 0;
	}
	_cur_log |= (1 << d);
	return 0;
}

static int bdos_resbdos(z80_sys_t *sys, bdos_params_t *parm) {
	_curdma = BDOS_DEFDMA;
	_curdma_real = mem_get_addr(&sys->mem, _curdma);
	_curdsk = 0;
	_curusr = 0;
	_curchr = 0;
	_cur_ro = 0;
	_cur_log = 0x0001;	// drive A: always present
	if (_cur_search.dir) {
		closedir(_cur_search.dir);
		_cur_search.dir = NULL;
	}
	uint16_t dpb = BDOS_START + OFFSET_OF(bdos_page_t,dpb);
	uint8_t *m = mem_get_addr(&sys->mem, dpb);
	memcpy(m, &_cur_dpb, sizeof(_cur_dpb));
	return 0;
}

static int bdos_const(z80_sys_t *sys, bdos_params_t *parm) {
	int rc = bios_const(sys, parm);
	if (rc) return rc;
	parm->hl = (parm->a != 0);
	return 0;
}

static int bdos_getver(z80_sys_t *sys, bdos_params_t *parm) {
	parm->hl = 0x0022;
	return 0;
}

static int bdos_user(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t e = parm->de & 0x00ff;
	if (e == 0xff) {
		parm->hl = _curusr;
		return 0;
	}
	_curusr = e & 0x1f;	// CP/M 3 requires 0x0f
	return 0;
}

static int bdos_bufin(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m;
	if (parm->de) {
		m = mem_get_addr(&sys->mem, parm->de);
	} else {
		m = _curdma_real;
	}
	uint8_t *b = m + 2;
	uint8_t mx = m[0];
	uint8_t nc = m[1];
	nc = 0;
	char *s = intr_fgets(sys, b + nc, mx - nc, stdin);
	if (sys->run == 0) return 0;
	if (!s) {
		parm->hl = 0xff;
		return 0;
	}
	nc = strlen(s);
	if (nc && s[nc - 1] == '\n') s[--nc] = '\0';
	parm->hl = nc;
	m[1] = nc;
	return 0;
}

static int bdos_open(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	int rc = _open_file(fcb);
	if (rc < 0) {
		parm->hl = 0xff;
		return 0;
	}
	return 0;
}

static int bdos_close(z80_sys_t *sys, bdos_params_t *parm) {
	int rc;
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	if (!fcb->s1[1]) {
		// special truncate for SUBMIT (CCP)
		off_t len = fcb->rc;
		len *= 128;
		rc = ftruncate(fcb->fd, len);
		if (rc < 0) {
			perror("ftruncate");
		}
	}
	rc = close(fcb->fd);
	fcb->fd = 0;
	if (rc < 0) {
		parm->hl = 0xff;
		return 0;
	}
	return 0;
}

static int bdos_search_next(z80_sys_t *sys, bdos_params_t *parm) {
	char *name = _do_search(&_cur_search);
	if (!name) {
		parm->hl = 255;
		return 0;
	}
	_copyout_dir(_curdma_real, name);
	if (_cur_search.ext == '?') {
		// return size of file... by some definition...
		// create fake info in FCB to fool caller into
		// thinking the file is of about the right size.
		fcb_impl_t *fcb = (fcb_impl_t *)_curdma_real;
		off_t len = _cur_search.size;
		len = (len + 127) / 128; // num records
		fcb->ex = (len >> 7) & 0x1f;	// max size < 512K
		fcb->rc = len & 0x7f;
		len = (len + _cur_dpb.blm) >> _cur_dpb.bsh;
		if (len > 16) len = 16;
		while (len) {
			--len;
			fcb->d0[len] = len + 1;
		}
	}
	parm->hl = 0;	// always index 0
	return 0;
}

static int bdos_search(z80_sys_t *sys, bdos_params_t *parm) {

	if (_cur_search.dir) {
		closedir(_cur_search.dir);
		_cur_search.dir = NULL;
	}
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
if (fcb->drv == '?') fprintf(stderr, "Search with drv = %02x\n", fcb->ex);
	int rc = _start_search(fcb, &_cur_search);
	if (rc < 0) {
		// no drive present...
		_cur_log &= ~(1 << _cur_search.drv);
		parm->hl = 255;
		return 0;
	}
	_cur_log |= (1 << _cur_search.drv);
	return bdos_search_next(sys, parm);
}

static int bdos_delete(z80_sys_t *sys, bdos_params_t *parm) {
	search_t era;
	char buf[256];
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	int rc = _start_search(fcb, &era);
	if (rc < 0) {
		_cur_log &= ~(1 << era.drv);
		parm->hl = 255;
		return 0;
	}
	_cur_log |= (1 << era.drv);
	rc = 0;
	char *name;
	while ((name = _do_search(&era)) != NULL) {
		++rc;
		sprintf(buf, "%s/%c:/%s", CPM_PREFIX, era.drv + 'A', name);
		unlink(buf);
	}
	if (rc == 0) {
		parm->hl = 255;
		return 0;
	}
	return 0;
}

static int bdos_readseq(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	if (!fcb->fd) {
		parm->hl = 9;
		return 0;
	}
	if (fcb->cr != fcb->s1[0]) {
		fcb->s1[0] = fcb->cr;
		off_t len = fcb->cr;
		len *= 128;
		lseek(fcb->fd, len, SEEK_SET);
	}
	int rc = read(fcb->fd, _curdma_real, 128);
	if (rc < 0) {
		parm->hl = 255;
		return 0;
	}
	if (rc == 0) {
		parm->hl = 1;
		return 0;
	}
	// detect media change?
	return 0;
}

static int bdos_writeseq(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	if (!fcb->fd) {
		parm->hl = 9;
		return 0;
	}
if (0) {
int x;
for (x = 0; x < 128; ++x) {
	if (_curdma_real[x] < ' ') {
		fputc('^',stderr);
		fputc(_curdma_real[x]+'@',stderr);
	} else {
		fputc(_curdma_real[x],stderr);
	}
}
fputc('\n',stderr);
}
	int rc = write(fcb->fd, _curdma_real, 128);
	if (rc < 0) {
		parm->hl = 255;
		return 0;
	}
	if (rc == 0) {
		parm->hl = 1;
		return 0;
	}
	// fcb->s1[1] = 0; // don't need this here... 
	// detect media change?
	return 0;
}

static int bdos_create(z80_sys_t *sys, bdos_params_t *parm) {
	char buf[16];
	char path[256];
	uint8_t d;
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	_get_fn(buf, fcb);
	d = fcb->drv;
	if (!d) {
		d = _curdsk;
	} else {
		--d;
	}
	uint8_t u = _curusr;
	if (u) {
		sprintf(path, "%s/%c:/%d:%s", CPM_PREFIX, d + 'A', u, buf);
	} else {
		sprintf(path, "%s/%c:/%s", CPM_PREFIX, d + 'A', buf);
	}
	int fd = open(path, O_RDWR|O_CREAT|O_EXCL, 0666);
	if (fd < 0) {
		if (errno == ENOENT) _cur_log &= ~(1 << d);
		parm->hl = 0xff;
		return 0;
	}
	_cur_log |= (1 << d);
	fcb->ex = 0;
	fcb->rc = 0;
	fcb->fd = fd;
	fcb->s1[1] = 0x80;
	return 0;
}

static int bdos_rename(z80_sys_t *sys, bdos_params_t *parm) {
	char buf[32];
	char old[256];
	char new[256];
	uint8_t d;
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	_get_fn(buf, fcb);
	d = fcb->drv;
	if (!d) {
		d = _curdsk;
	} else {
		--d;
	}
	uint8_t u = _curusr;
	if (u) {
		sprintf(old, "%s/%c:/%d:%s", CPM_PREFIX, d + 'A', u, buf);
	} else {
		sprintf(old, "%s/%c:/%s", CPM_PREFIX, d + 'A', buf);
	}
	fcb = (fcb_impl_t *)(m + 16);
	_get_fn(buf, fcb);
	if (u) {
		sprintf(new, "%s/%c:/%d:%s", CPM_PREFIX, d + 'A', u, buf);
	} else {
		sprintf(new, "%s/%c:/%s", CPM_PREFIX, d + 'A', buf);
	}
	int rc = rename(old, new);
	if (rc < 0) {
		if (errno == ENOENT) _cur_log &= ~(1 << d);
		// todo: decode errno...
		parm->hl = 0xff;
		return 0;
	}
	_cur_log |= (1 << d);
	return 0;
}

static int bdos_curdsk(z80_sys_t *sys, bdos_params_t *parm) {
	parm->hl = _curdsk;
	return 0;
}

static int bdos_rread(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	if (!fcb->fd) {
		parm->hl = 9;
		return 0;
	}
	_seek_file(fcb);
	int rc = read(fcb->fd, _curdma_real, 128);
	_seek_file(fcb);
	if (rc < 0) {
		parm->hl = 255;
		return 0;
	}
	if (rc == 0) {
		parm->hl = 1;
		return 0;
	}
	// detect media change?
	return 0;
}

static int bdos_rwrite(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	if (!fcb->fd) {
		parm->hl = 9;
		return 0;
	}
	_seek_file(fcb);
	int rc = write(fcb->fd, _curdma_real, 128);
	_seek_file(fcb);
	if (rc < 0) {
		parm->hl = 255;
		return 0;
	}
	if (rc == 0) {
		parm->hl = 1;
		return 0;
	}
	// detect media change?
	return 0;
}

static int bdos_setrr(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	if (!fcb->fd) {
		parm->hl = 9;
		return 0;
	}
	off_t r = lseek(fcb->fd, (off_t)0, SEEK_CUR);
	r = (r + 127) / 128;
	fcb->rr[0] = r & 0x0ff;
	r >>= 8;
	fcb->rr[1] = r & 0x0ff;
	r >>= 8;
	fcb->rr[2] = r & 0x0ff;
	return 0;
}

static int bdos_filesize(z80_sys_t *sys, bdos_params_t *parm) {
	uint8_t *m = mem_get_addr(&sys->mem, parm->de);
	fcb_impl_t *fcb = (fcb_impl_t *)m;
	if (!fcb->fd) {
		int rc = _open_file(fcb);
		if (rc < 0) {
			parm->hl = 255;
			return 0;
		}
	}
	struct stat stb;
	int rc = fstat(fcb->fd, &stb);
	if (rc < 0) {
		parm->hl = 255;
		return 0;
	}
	off_t r = stb.st_size;
	r = (r + 127) / 128;
	fcb->rr[0] = r & 0x0ff;
	r >>= 8;
	fcb->rr[1] = r & 0x0ff;
	r >>= 8;
	fcb->rr[2] = r & 0x0ff;
	return 0;
}

bdos_func_t *bdos_funcs[256] = {
[0] = bdos_reset,	// system reset
[1] = bdos_conin,	// console input
[2] = bdos_conout,	// console output
[3] = _bdos_unimpl,	// reader input
[4] = _bdos_unimpl,	// punch output
[5] = _bdos_unimpl,	// list output
[6] = _bdos_inval,	// direct console I/O
[7] = bdos_getiob,	// get iobyte
[8] = bdos_setiob,	// put iobyte
[9] = bdos_msgout,	// print string on console
[10] = bdos_bufin,	// read console buffer
[11] = bdos_const,	// get console status
[12] = bdos_getver,	// get version
[13] = bdos_resbdos,	// reset disk system
[14] = bdos_seldsk,	// select disk
[15] = bdos_open,	// open file
[16] = bdos_close,	// close file
[17] = bdos_search,	// search first
[18] = bdos_search_next,// search next
[19] = bdos_delete,	// delete file
[20] = bdos_readseq,	// read sequential
[21] = bdos_writeseq,	// write sequential
[22] = bdos_create,	// make file
[23] = bdos_rename,	// rename file
[24] = bdos_getlog,	// get login vector
[25] = bdos_curdsk,	// get current disk
[26] = bdos_setdma,	// set DMA addr
[27] = bdos_getall,	// get allocation addr
[28] = bdos_prot,	// write protext disk
[29] = bdos_getro,	// get R/O vector
[30] = _bdos_unimpl,	// set file attributes
[31] = bdos_getdpb,	// get dpb
[32] = bdos_user,	// set/get user code
[33] = bdos_rread,	// read random
[34] = bdos_rwrite,	// write random
[35] = bdos_filesize,	// compute file size
[36] = bdos_setrr,	// set random record
[37] = bdos_resdrv,	// reset drive
[38] = _bdos_inval,	// 
[39] = _bdos_inval,	// 
[40] = bdos_rwrite,	// write random w/zerofill - same as write random
};

void bdos_impl_init(z80_sys_t *sys) {
	uint8_t *m = mem_get_addr(&sys->mem, BDOS_ENTRY);
	*m++ = 0xc9;	// after trap, return here...
	*m++ = BDOS_START & 0x00ff;
	*m++ = BDOS_START >> 8;

	// loop back to bdos in case some tries direct call
	m = mem_get_addr(&sys->mem, BDOS_START);
	*m++ = 0xc3;
	*m++ = BDOS_ENTRY & 0x00ff;
	*m++ = BDOS_ENTRY >> 8;

	// give appearances of a BIOS, too
	m = mem_get_addr(&sys->mem, 0x0000);
	*m++ = 0xc9;	// after trap, return here...
	*m++ = (BIOS_START + 3) & 0x00ff;
	*m++ = (BIOS_START + 3) >> 8;

	// BIOS address are trapped, just like BDOS_ENTRY, but
	// some programs go so far as to pick up the routine address
	// from the JMP here, so we must ensure that not only
	// CALL BIOS_START+n works but also CALL *(BIOS_START+n+1).
	//
	uint16_t bios_pc = BIOS_START;
	m = mem_get_addr(&sys->mem, BIOS_START);
	int x;
	for (x = 0; x < 20; ++x) {
		*m++ = 0xc9;	// after trap, return here...
		*m++ = bios_pc & 0x00ff;
		*m++ = bios_pc >> 8;
		bios_pc += 3;
	}
	// rest is done by trapping any PC in the BIOS_START page.
}
