// $Id: z80-sim.c,v 1.11 2011-04-08 22:45:51 drmiller Exp $

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>

#include "z80_sys.h"

z80_sys_t sys;

static void set_intr(void) {
	struct sigaction sa;
	extern void sig_intr(int sig);
	sa.sa_handler = sig_intr;
	sa.sa_flags = SA_RESTART;
	sigemptyset(&sa.sa_mask);
	sigaction(SIGINT, &sa, NULL);
}

void sig_intr(int sig) {
	if (sys.intr && sys.intr(&sys, sig) == 0) {
		return;
	}
	sys.fault(&sys, "\nInterrupt");
	exit(1);
}

int main(int argc, char **argv) {
	sys_init(&sys);
	int x;
	uint16_t entry = -1, load = -1;
	char *pgm = NULL;
	uint16_t ccp_entry = -1, ccp_load = -1;
	char *ccp = NULL;
	int interact = 0;

	extern char *optarg;
	extern int optind, opterr, optopt;

	while ((x = getopt(argc, argv, "C:E:e:iL:l:p:t")) != EOF) {
		switch(x) {
#ifdef USE_BDOS
		case 'C':
			ccp = optarg;
			break;
		case 'E':
			ccp_entry = strtoul(optarg, NULL, 0);
			break;
		case 'L':
			ccp_load = strtoul(optarg, NULL, 0);
			break;
		case 'i':
			interact = 1;
			break;
#endif // USE_BDOS
		case 'e':
			entry = strtoul(optarg, NULL, 0);
			break;
		case 'l':
			load = strtoul(optarg, NULL, 0);
			break;
		case 'p':
			pgm = optarg;
			break;
#ifdef TRACE
		case 't':
			sys.trace = 1;
			break;
#endif // TRACE
		default:
			fprintf(stderr, "Usage: "
					"%s [-p program [-l load-addr] [-e entry]]\n",
					argv[0]);
			exit(1);
			break;
		}
	}
	if (load == (uint16_t)-1) {
		load = 0x0100;
	}
	if (entry == (uint16_t)-1) {
		entry = load;
	}
	if (ccp) {
#ifdef USE_BDOS
		if (ccp_load == (uint16_t)-1) {
			ccp_load = 0x0100;
		}
		if (ccp_entry == (uint16_t)-1) {
			ccp_entry = ccp_load;
		}
		sys_setccp(&sys, ccp, ccp_load, ccp_entry);
#endif // USE_BDOS
	} else if (pgm == NULL) {
		pgm = "cpm_program";
	}

	set_intr();

	if (pgm) {
		sys_loadpgm(&sys, pgm, load, entry);
	}

	sys.run = (interact ? 0 : 1);
	if (ccp) {
#ifdef USE_BDOS
		sys_runccp(&sys);
#endif // USE_BDOS
	} else {
		sys_go(&sys, entry);
	}
	return 0;
}
