// $Id: z80_alu.c,v 1.9 2011-04-06 21:44:04 drmiller Exp $

#include "z80_alu.h"

uint16_t add16(uint32_t a, uint32_t b, z80_psw_t *f) {
	uint32_t r = a + b;
	f->psw_CY = ((r & 0x00010000) != 0);
	f->psw_NN = 0;
	return r;
}

uint16_t adc16(uint32_t a, uint32_t b, z80_psw_t *f) {
	uint32_t r = a + b + f->psw_CY;

	int nc = ((((a & 0x007fff) + (b & 0x007fff) + f->psw_CY) & 0x008000) != 0);
	int cy = ((((a & 0x00ffff) + (b & 0x00ffff) + f->psw_CY) & 0x010000) != 0);

	f->psw_CY = cy;
	f->psw_PV = (cy != nc);
	f->psw_ZR = ((r & 0x0000ffff) == 0);
	f->psw_SG = ((r & 0x00008000) != 0);
	f->psw_NN = 0;
	return r;
}

uint16_t sbc16(uint32_t a, uint32_t b, z80_psw_t *f) {
	uint32_t r = a - b - f->psw_CY;

	int nc = ((((a & 0x007fff) - (b & 0x007fff) - f->psw_CY) & 0x008000) != 0);
	int cy = ((((a & 0x00ffff) - (b & 0x00ffff) - f->psw_CY) & 0x010000) != 0);

	f->psw_CY = cy;
	f->psw_PV = (cy != nc);
	f->psw_ZR = ((r & 0x0000ffff) == 0);
	f->psw_SG = ((r & 0x00008000) != 0);
	f->psw_NN = 1;
	return r;
}

uint16_t add8(uint16_t a, uint16_t b, z80_psw_t *f) {
	uint16_t r = a + b;

	int hc = ((((a & 0x000f) + (b & 0x000f)) & 0x0010) != 0);
	int nc = ((((a & 0x007f) + (b & 0x007f)) & 0x0080) != 0);
	int cy = ((((a & 0x00ff) + (b & 0x00ff)) & 0x0100) != 0);

	f->psw_CY = cy;
	f->psw_HC = hc;
	f->psw_PV = (cy != nc);
	f->psw_SG = ((r & 0x0080) != 0);
	f->psw_ZR = ((r & 0x00ff) == 0);
	f->psw_NN = 0;
	return r;
}

uint16_t adc8(uint16_t a, uint16_t b, z80_psw_t *f) {
	uint16_t r = a + b + f->psw_CY;

	int hc = ((((a & 0x000f) + (b & 0x000f) + f->psw_CY) & 0x0010) != 0);
	int nc = ((((a & 0x007f) + (b & 0x007f) + f->psw_CY) & 0x0080) != 0);
	int cy = ((((a & 0x00ff) + (b & 0x00ff) + f->psw_CY) & 0x0100) != 0);

	f->psw_CY = cy;
	f->psw_HC = hc;
	f->psw_PV = (cy != nc);
	f->psw_SG = ((r & 0x0080) != 0);
	f->psw_ZR = ((r & 0x00ff) == 0);
	f->psw_NN = 0;
	return r;
}

uint16_t sub8(uint16_t a, uint16_t b, z80_psw_t *f) {
	uint16_t r = a - b;

	int hc = ((((a & 0x000f) - (b & 0x000f)) & 0x0010) != 0);
	int nc = ((((a & 0x007f) - (b & 0x007f)) & 0x0080) != 0);
	int cy = ((((a & 0x00ff) - (b & 0x00ff)) & 0x0100) != 0);

	f->psw_CY = cy;
	f->psw_HC = hc;
	f->psw_PV = (cy != nc);
	f->psw_SG = ((r & 0x0080) != 0);
	f->psw_ZR = ((r & 0x00ff) == 0);
	f->psw_NN = 1;
	return r;
}

uint16_t sbb8(uint16_t a, uint16_t b, z80_psw_t *f) {
	uint16_t r = a - b - f->psw_CY;

	int hc = ((((a & 0x000f) - (b & 0x000f) - f->psw_CY) & 0x0010) != 0);
	int nc = ((((a & 0x007f) - (b & 0x007f) - f->psw_CY) & 0x0080) != 0);
	int cy = ((((a & 0x00ff) - (b & 0x00ff) - f->psw_CY) & 0x0100) != 0);

	f->psw_CY = cy;
	f->psw_HC = hc;
	f->psw_PV = (cy != nc);
	f->psw_SG = ((r & 0x0080) != 0);
	f->psw_ZR = ((r & 0x00ff) == 0);
	f->psw_NN = 1;
	return r;
}

uint8_t parity[256] = {
	1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,	//   0- 15     0000-    1111
	0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,	//  16- 31    10000-   11111
	0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,	//  32- 47   100000-  101111
	1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,	//  48- 63   110000-  111111
	0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,	//  64- 79  1000000- 1001111
	1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,	//  80- 95  1010000- 1011111
	1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,	//  96-111  1100000- 1101111
	0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,	// 112-127  1110000- 1111111

	0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,	// 128-143 10000000-10001111
	1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,	// 144-159 10010000-10011111
	1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,	// 160-175 10100000-10101111
	0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,	// 176-191 10110000-10111111
	1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,	// 192-207 11000000-11001111
	0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,	// 208-223 11010000-11011111
	0,1,1,0,1,0,0,1,1,0,0,1,0,1,1,0,	// 224-239 11100000-11101111
	1,0,0,1,0,1,1,0,0,1,1,0,1,0,0,1,	// 240-255 11110000-11111111
};

uint16_t and8(uint16_t a, uint16_t b, z80_psw_t *f) {
	uint8_t r = a & b;

	f->psw_CY = 0;
	f->psw_HC = 0;
	f->psw_PV = parity[r];
	f->psw_SG = ((r & 0x0080) != 0);
	f->psw_ZR = ((r & 0x00ff) == 0);
	f->psw_NN = 1;
	return r;
}

uint16_t xor8(uint16_t a, uint16_t b, z80_psw_t *f) {
	uint8_t r = a ^ b;

	f->psw_CY = 0;
	f->psw_HC = 0;
	f->psw_PV = parity[r];
	f->psw_SG = ((r & 0x0080) != 0);
	f->psw_ZR = ((r & 0x00ff) == 0);
	f->psw_NN = 1;
	return r;
}

uint16_t or8(uint16_t a, uint16_t b, z80_psw_t *f) {
	uint8_t r = a | b;

	f->psw_CY = 0;
	f->psw_HC = 0;
	f->psw_PV = parity[r];
	f->psw_SG = ((r & 0x0080) != 0);
	f->psw_ZR = ((r & 0x00ff) == 0);
	f->psw_NN = 1;
	return r;
}

uint16_t rlc8a(uint16_t a, z80_psw_t *f) {
	uint16_t v = a;
	v <<= 1;
	uint8_t cy = ((v & 0x100) >> 8);
	v |= cy;
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_NN = 0;
	return v;
}

uint16_t rrc8a(uint16_t a, z80_psw_t *f) {
	uint16_t v = a << 7;   // 0AAAAAAA|A0000000
	uint8_t cy = ((v & 0x080) >> 7);
	v = (v >> 8) | (v & 0x0080); 
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_NN = 0;
	return v;
}
	
uint16_t ral8a(uint16_t a, z80_psw_t *f) {
	uint16_t v = a;
	v <<= 1; 
	uint8_t cy = ((v & 0x100) >> 8);
	v |= f->psw_CY;
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_NN = 0;
	return v;
}

uint16_t rar8a(uint16_t a, z80_psw_t *f) {
	uint16_t v = a << 7;   // 0AAAAAAA|A0000000
	uint8_t cy = ((v & 0x080) >> 7);
	v = (v >> 8) | (f->psw_CY << 7);
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_NN = 0;
	return v;
}

uint16_t rlc8r(uint16_t a, z80_psw_t *f) {
	uint16_t v = a;
	v <<= 1;
	uint8_t cy = ((v & 0x100) >> 8);
	v |= cy;
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x00ff];
	f->psw_SG = ((v & 0x0080) != 0);
	f->psw_ZR = ((v & 0x00ff) == 0);
	f->psw_NN = 0;
	return v;
}

uint16_t rrc8r(uint16_t a, z80_psw_t *f) {
	uint16_t v = a << 7;   // 0AAAAAAA|A0000000
	uint8_t cy = ((v & 0x080) >> 7);
	v = (v >> 8) | (v & 0x0080);
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x00ff];
	f->psw_SG = ((v & 0x0080) != 0);
	f->psw_ZR = ((v & 0x00ff) == 0);
	f->psw_NN = 0;
	return v;
}

uint16_t ral8r(uint16_t a, z80_psw_t *f) {
	uint16_t v = a;
	v <<= 1; 
	uint8_t cy = ((v & 0x100) >> 8);
	v |= f->psw_CY;
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x00ff];
	f->psw_SG = ((v & 0x0080) != 0);
	f->psw_ZR = ((v & 0x00ff) == 0);
	f->psw_NN = 0;
	return v;
}

uint16_t rar8r(uint16_t a, z80_psw_t *f) {
	uint16_t v = a << 7;   // 0AAAAAAA|A0000000
	uint8_t cy = ((v & 0x080) >> 7);
	v = (v >> 8) | (f->psw_CY << 7);
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x00ff];
	f->psw_SG = ((v & 0x0080) != 0);
	f->psw_ZR = ((v & 0x00ff) == 0);
	f->psw_NN = 0;
	return v;
}

uint16_t srl8r(uint16_t a, z80_psw_t *f) {
	uint16_t v = a;
	uint8_t cy = (v & 0x0001);
	v >>= 1;
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x00ff];
	f->psw_SG = ((v & 0x0080) != 0);
	f->psw_ZR = ((v & 0x00ff) == 0);
	f->psw_NN = 0;
	return v;
}

uint16_t sra8r(uint16_t a, z80_psw_t *f) {
	uint16_t v = a;
	uint8_t cy = (v & 0x0001);
	v = (v >> 1) | (v & 0x0080);
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x00ff];
	f->psw_SG = ((v & 0x0080) != 0);
	f->psw_ZR = ((v & 0x00ff) == 0);
	f->psw_NN = 0;
	return v;
}

uint16_t sla8r(uint16_t a, z80_psw_t *f) {
	uint16_t v = a;
	uint8_t cy = ((v & 0x0080) >> 7);
	v <<= 1;
	f->psw_CY = cy;
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x00ff];
	f->psw_SG = ((v & 0x0080) != 0);
	f->psw_ZR = ((v & 0x00ff) == 0);
	f->psw_NN = 0;
	return v;
}

uint32_t rld8r(uint32_t a, z80_psw_t *f) {
	uint32_t v = a;	// 00000000|00000000|aaaaAAAA|mmmmMMMM
	v = (v & 0x0000f000) | ((v & 0x000000ff) << 4) | ((v & 0x00000f00) >> 8);
			// 00000000|00000000|aaaammmm|MMMMAAAA
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x000000ff];
	f->psw_SG = ((v & 0x00000080) != 0);
	f->psw_ZR = ((v & 0x000000ff) == 0);
	f->psw_NN = 0;
	return v;
}

uint32_t rrd8r(uint32_t a, z80_psw_t *f) {
	uint32_t v = a;	// 00000000|00000000|aaaaAAAA|mmmmMMMM
	v = (v & 0x0000f000) | ((v & 0x00000ff0) >> 4) | ((v & 0x0000000f) << 8);
			// 00000000|00000000|aaaaMMMM|AAAAmmmm
	f->psw_HC = 0;
	f->psw_PV = parity[v & 0x000000ff];
	f->psw_SG = ((v & 0x00000080) != 0);
	f->psw_ZR = ((v & 0x000000ff) == 0);
	f->psw_NN = 0;
	return v;
}
