#ifndef __z80_alu_h__
#define __z80_alu_h__

// $Id: z80_alu.h,v 1.4 2011-04-03 17:22:21 drmiller Exp $
#include "z80_cpu.h"

// NOTE: these operations are performed on their input parameters ONLY,
// with the result returned to caller. The only CPU register affected is the FLAGS.
// The caller retrieves the input values as appropriate for their use case,
// and stores the result value similarly.

// 16-but "double add", update CY only
extern uint16_t add16(uint32_t a, uint32_t b, z80_psw_t *f);

// add 8-bit, update all flags (PV = overflow)
extern uint16_t add8(uint16_t a, uint16_t b, z80_psw_t *f);

// add w/CY 8-bit, update all flags (PV = overflow)
extern uint16_t adc8(uint16_t a, uint16_t b, z80_psw_t *f);

// subtract 8-bit, update all flags (PV = overflow)
extern uint16_t sub8(uint16_t a, uint16_t b, z80_psw_t *f);

// subtract w/CY (borrow) 8-bit, update all flags (PV = overflow)
extern uint16_t sbb8(uint16_t a, uint16_t b, z80_psw_t *f);

// bit-wise AND, update all flags (PV = parity)
extern uint16_t and8(uint16_t a, uint16_t b, z80_psw_t *f);

// bit-wise XOR, update all flags (PV = parity)
extern uint16_t xor8(uint16_t a, uint16_t b, z80_psw_t *f);

// bit-wise OR, update all flags (PV = parity)
extern uint16_t or8(uint16_t a, uint16_t b, z80_psw_t *f);

extern uint16_t rlc8a(uint16_t a, z80_psw_t *f);
extern uint16_t rlc8r(uint16_t a, z80_psw_t *f);
extern uint16_t rrc8a(uint16_t a, z80_psw_t *f);
extern uint16_t rrc8r(uint16_t a, z80_psw_t *f);
extern uint16_t ral8a(uint16_t a, z80_psw_t *f);
extern uint16_t ral8r(uint16_t a, z80_psw_t *f);
extern uint16_t rar8a(uint16_t a, z80_psw_t *f);
extern uint16_t rar8r(uint16_t a, z80_psw_t *f);
extern uint16_t srl8r(uint16_t a, z80_psw_t *f);
extern uint16_t sra8r(uint16_t a, z80_psw_t *f);
extern uint16_t sla8r(uint16_t a, z80_psw_t *f);
// result: low 16 to (HL), high 16 (8) to A
extern uint32_t rld8r(uint32_t a, z80_psw_t *f);
extern uint32_t rrd8r(uint32_t a, z80_psw_t *f);

#endif // __z80_alu_h__
