// $Id: z80_exec.c,v 1.12 2011-04-08 23:11:31 drmiller Exp $

#include <stdlib.h>

#include "z80_cpu.h"
#include "z80_decode.h"
#include "z80_decode2.h"
#include "z80_decode3.h"

#undef TRACE_CYCLES

#ifdef TRACE
inline uint64_t tsc() {
	uint64_t tsc;
	__asm__ __volatile__ ("rdtsc" : "=A" (tsc));
	return tsc;
}

extern int diz80(char *buf, uint8_t *t);
#endif // TRACE

// BC DE HL SP
uint16_t *get_rp(z80_sys_t *sys, int rp) {
	if (rp == 3) {
		return &sys->cpu.sp;
	} else {
		return &sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f & 1].xx[rp];
	}
}

// BC DE HL AF - push/pop etc
uint16_t *get_rp2(z80_sys_t *sys, int rp) {
	if (rp == 3) {
		return &sys->cpu.psw.af[sys->cpu.psw.af_f & 1].xx;
	} else {
		return &sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f & 1].xx[rp];
	}
}

// BC DE IX/IY SP
uint16_t *get_rp3(z80_sys_t *sys, int rp, int ix) {
	if (rp == 3) {
		return &sys->cpu.sp;
	} else if (rp == 2) {
		return &sys->cpu.ix[ix];
	} else {
		return &sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f & 1].xx[rp];
	}
}

// B C D E H L m A
uint8_t *get_gp(z80_sys_t *sys, int gp) {
	if (gp == 6) {
		uint16_t hl = *(get_rp(sys, rp_HL));
		sys->cpu.cycles += 3; // maybe not exactly...
		return mem_get_addr(&sys->mem, hl);
	} else if (gp == 7) {
		return &sys->cpu.psw.af[sys->cpu.psw.af_f & 1].af[af_A];
	} else {
		return &sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f & 1].gp[gp ^ 1];
	}
}

uint8_t *get_gp_A(z80_sys_t *sys) {
	return &sys->cpu.psw.af[sys->cpu.psw.af_f & 1].af[af_A];
}

z80_psw_t *get_gp_F(z80_sys_t *sys) {
	return (z80_psw_t *)&sys->cpu.psw.af[sys->cpu.psw.af_f & 1].af[af_F];
}

char *get_psw_str(z80_sys_t *sys, z80_psw_t *f) {
	static char buf[16];
	buf[0] = f->psw_SG ? 'S' : 's';
	buf[1] = f->psw_ZR ? 'Z' : 'z';
	buf[2] = '-';
	buf[3] = f->psw_HC ? 'H' : 'h';
	buf[4] = '-';
	buf[5] = f->psw_PV ? 'V' : 'v';
	buf[6] = f->psw_NN ? 'N' : 'n';
	buf[7] = f->psw_CY ? 'C' : 'c';
	buf[8] = '\0';
	return buf;
}

uint8_t fetch_opcode(z80_sys_t *sys, int off) {
	sys->cpu.cycles += 4;
	++sys->cpu.re;
	uint16_t a = sys->cpu.pc + off;
	return mem_get_byte(&sys->mem, a);
}

uint8_t fetch_imm8(z80_sys_t *sys, int off) {
	sys->cpu.cycles += 3;
	uint16_t a = sys->cpu.pc + off;
	return mem_get_byte(&sys->mem, a);
}

uint16_t fetch_imm16(z80_sys_t *sys, int off) {
	sys->cpu.cycles += 6;
	uint16_t a = sys->cpu.pc + off;
	return mem_get_word(&sys->mem, a);
}

void ill_instr(z80_sys_t *sys) {
	uint16_t pc = sys->cpu.pc;
	fprintf(stderr, "Illegal instruction %04x: %02x %02x %02x %02x\n", pc,
		mem_get_byte(&sys->mem, pc + 0),
		mem_get_byte(&sys->mem, pc + 1),
		mem_get_byte(&sys->mem, pc + 2),
		mem_get_byte(&sys->mem, pc + 3));
	fprintf(stderr, "Terminating at cycle %lld\n", sys->cpu.cycles);
	exit(1);
}

void sys_exec(z80_sys_t *sys) {
#ifdef TRACE
	uint16_t z80_pc = sys->cpu.pc;
#ifdef TRACE_CYCLES
	uint64_t nat_tsc = 0;
	uint64_t z80_cy = 0;
	if (sys->trace) {
		nat_tsc = tsc();
		z80_cy = sys->cpu.cycles;
	}
#endif // TRACE_CYCLES
#endif // TRACE
	uint8_t op1 = fetch_opcode(sys, 0);
	instr_t *f = primary[op1];
	if (!f) { ill_instr(sys); }
	f(sys, op1, 1);
#ifdef TRACE
	if (sys->trace) {
		uint8_t *m;
		int c, x;
#ifdef TRACE_CYCLES
		nat_tsc = tsc() - nat_tsc;
		z80_cy = sys->cpu.cycles - z80_cy;
#endif // TRACE_CYCLES
		char buf[64];
		m = mem_get_addr(&sys->mem, z80_pc);
		diz80(buf, m);
		fprintf(sys->trc_fp, "TRACE: %04x [ %-15s ] "
#ifdef TRACE_CYCLES
				"%2lld z80 cycles, %5lld native cycles\n",
				z80_pc, buf, z80_cy, nat_tsc);
#else // !TRACE_CYCLES
				"%02x [%s] %04x %04x %04x %04x %04x\n",
				z80_pc, buf,
				*(get_gp_A(sys)),
				get_psw_str(sys, get_gp_F(sys)),
				*(get_rp(sys, rp_BC)),
				*(get_rp(sys, rp_DE)),
				*(get_rp(sys, rp_HL)),
				sys->cpu.ix[ix_X],
				sys->cpu.ix[ix_Y]);
#endif // !TRACE_CYCLES
	}
#endif // TRACE
}

void opcode_CB(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t op2 = fetch_opcode(sys, off);
	instr2_t *f = sec_CB[op2];
	if (!f) { ill_instr(sys); }
	f(sys, op1, op2, off + 1);
}

void opcode_DD(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t op2 = fetch_opcode(sys, off);
	instr2_t *f = sec_DD[op2];
	if (!f) { ill_instr(sys); }
	f(sys, op1, op2, off + 1);
}

void opcode_ED(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t op2 = fetch_opcode(sys, off);
	instr2_t *f = sec_ED[op2];
	if (!f) { ill_instr(sys); }
	f(sys, op1, op2, off + 1);
}

void opcode_FD(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t op2 = fetch_opcode(sys, off);
	instr2_t *f = sec_FD[op2];
	if (!f) { ill_instr(sys); }
	f(sys, op1, op2, off + 1);
}

void opcode_DD_CB(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {	// also FD_CB...
	uint8_t op3 = fetch_opcode(sys, off + 1);
	instr3_t *f = ter_DD_CB[op3];
	if (!f) { ill_instr(sys); }
	f(sys, op1, op2, op3, off + 2);
}
