// $Id: z80_instr.c,v 1.16 2011-04-06 21:44:04 drmiller Exp $

// one-byte opcode instructions
#include "z80_instr.h"
#include "z80_exec.h"
#include "z80_alu.h"

void no_op_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	sys->cpu.pc += off;
}

void add_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint8_t *r = get_gp(sys, (op1 & 0007));
	z80_psw_t *f = get_gp_F(sys);

	*a = add8(*a, *r, f);
	sys->cpu.pc += off;
}

void adc_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint8_t *r = get_gp(sys, (op1 & 0007));
	z80_psw_t *f = get_gp_F(sys);

	*a = adc8(*a, *r, f);
	sys->cpu.pc += off;
}

void sub_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint8_t *r = get_gp(sys, (op1 & 0007));
	z80_psw_t *f = get_gp_F(sys);

	*a = sub8(*a, *r, f);
	sys->cpu.pc += off;
}

void sbb_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint8_t *r = get_gp(sys, (op1 & 0007));
	z80_psw_t *f = get_gp_F(sys);

	*a = sbb8(*a, *r, f);
	sys->cpu.pc += off;
}

void and_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint8_t *r = get_gp(sys, (op1 & 0007));
	z80_psw_t *f = get_gp_F(sys);

	*a = and8(*a, *r, f);
	sys->cpu.pc += off;
}

void xor_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint8_t *r = get_gp(sys, (op1 & 0007));
	z80_psw_t *f = get_gp_F(sys);

	*a = xor8(*a, *r, f);
	sys->cpu.pc += off;
}

void or_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint8_t *r = get_gp(sys, (op1 & 0007));
	z80_psw_t *f = get_gp_F(sys);

	*a = or8(*a, *r, f);
	sys->cpu.pc += off;
}

void cmp_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint8_t *r = get_gp(sys, (op1 & 0007));
	z80_psw_t *f = get_gp_F(sys);

	(void)sub8(*a, *r, f);
	sys->cpu.pc += off;
}

void inc_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *r = get_gp(sys, (op1 & 0070) >> 3);
	z80_psw_t *f = get_gp_F(sys);

	int cy = f->psw_CY;

	*r = add8(*r, 1, f);
	 f->psw_CY = cy;
	if ((op1 & 0007) == 6) sys->cpu.cycles += 4;
	sys->cpu.pc += off;
}

void dec_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *r = get_gp(sys, (op1 & 0070) >> 3);
	z80_psw_t *f = get_gp_F(sys);

	int cy = f->psw_CY;

	*r = sub8(*r, 1, f);
	 f->psw_CY = cy;
	if ((op1 & 0007) == 6) sys->cpu.cycles += 4;
	sys->cpu.pc += off;
}

void dad_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *r = get_rp(sys, (op1 & 0060) >> 4);
	uint16_t *h = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);

	*h = add16(*h, *r, f);
	sys->cpu.cycles += 7;
	sys->cpu.pc += off;
}

void di_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	sys->cpu.di = 1;
	sys->cpu.pc += off;
}

void ei_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	sys->cpu.di = 0;
	sys->cpu.pc += off;
}

void mov_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *d = get_gp(sys, (op1 & 0070) >> 3);
	uint8_t *s = get_gp(sys, (op1 & 0007));
	*d = *s;
	sys->cpu.pc += off;
}

void mvi_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *d = get_gp(sys, (op1 & 0070) >> 3);
	uint8_t n = fetch_imm8(sys, off);
	*d = n;
	sys->cpu.pc += off + 1;
}

void lxi_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *d = get_rp(sys, (op1 & 0060) >> 4);
	uint16_t n = fetch_imm16(sys, off);
	*d = n;
	sys->cpu.pc += off + 2;
}

void inx_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *d = get_rp(sys, (op1 & 0060) >> 4);
	*d += 1;
	sys->cpu.cycles += 2;
	sys->cpu.pc += off;
}

void dcx_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *d = get_rp(sys, (op1 & 0060) >> 4);
	*d -= 1;
	sys->cpu.cycles += 2;
	sys->cpu.pc += off;
}

void ldax_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t rr = *(get_rp(sys, (op1 & 0060) >> 4));
	uint8_t *a = get_gp_A(sys);
	*a = *(mem_get_addr(&sys->mem, rr));
	sys->cpu.cycles += 3;
	sys->cpu.pc += off;
}

void stax_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t rr = *(get_rp(sys, (op1 & 0060) >> 4));
	uint8_t *a = get_gp_A(sys);
	*(mem_get_addr(&sys->mem, rr)) = *a;
	sys->cpu.cycles += 3;
	sys->cpu.pc += off;
}

void lda_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t nn = fetch_imm16(sys, off);
	uint8_t *a = get_gp_A(sys);
	*a = *(mem_get_addr(&sys->mem, nn));
	sys->cpu.cycles += 3;
	sys->cpu.pc += off + 2;
}

void sta_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t nn = fetch_imm16(sys, off);
	uint8_t *a = get_gp_A(sys);
	*(mem_get_addr(&sys->mem, nn)) = *a;
	sys->cpu.cycles += 3;
	sys->cpu.pc += off + 2;
}

void lhld_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *hl = get_rp(sys, rp_HL);
	uint16_t nn = fetch_imm16(sys, off);
	uint16_t *t = (uint16_t *)mem_get_addr(&sys->mem, nn);
	*hl = *t;
	sys->cpu.cycles += 6;
	sys->cpu.pc += off + 2;
}

void shld_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *hl = get_rp(sys, rp_HL);
	uint16_t nn = fetch_imm16(sys, off);
	uint16_t *t = (uint16_t *)mem_get_addr(&sys->mem, nn);
	*t = *hl;
	sys->cpu.cycles += 6;
	sys->cpu.pc += off + 2;
}

void xchg_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *de = get_rp(sys, rp_DE);
	uint16_t *hl = get_rp(sys, rp_HL);
	uint16_t t = *de;
	*de = *hl;
	*hl = t;
	sys->cpu.pc += off;
}

void exx_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	sys->cpu.gp_regs.gp_f ^= 1;
	sys->cpu.pc += off;
}

void exaf_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	sys->cpu.psw.af_f ^= 1;
	sys->cpu.pc += off;
}

void xthl_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *hl = get_rp(sys, rp_HL);
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	uint16_t t = *hl;
	*hl = *ts;
	*ts = t;
	sys->cpu.cycles += 15;
	sys->cpu.pc += off;
}

void cpl_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	*a = ~(*a);
	z80_psw_t *f = get_gp_F(sys);
	f->psw_NN = 1;
	f->psw_HC = 1;
	sys->cpu.pc += off;
}

void ccf_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	f->psw_CY = !f->psw_CY;
	sys->cpu.pc += off;
}

void scf_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	f->psw_CY = 1;
	sys->cpu.pc += off;
}

void pchl_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *hl = get_rp(sys, rp_HL);
	sys->cpu.pc = *hl;
}

void sphl_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *hl = get_rp(sys, rp_HL);
	sys->cpu.sp = *hl;
	sys->cpu.pc += off;
}

void jmp_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t a = fetch_imm16(sys, off);
	sys->cpu.pc = a;
}

void jcc_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t a = fetch_imm16(sys, off);
	int cc = ((op1 & 070) >> 3);
	z80_psw_t f = *(get_gp_F(sys));
	int doit = 0;
	switch (cc) {
	case 0:
		doit = (!f.psw_ZR);
		break;
	case 1:
		doit = (f.psw_ZR);
		break;
	case 2:
		doit = (!f.psw_CY);
		break;
	case 3:
		doit = (f.psw_CY);
		break;
	case 4:
		doit = (!f.psw_PV);
		break;
	case 5:
		doit = (f.psw_PV);
		break;
	case 6:
		doit = (!f.psw_SG);
		break;
	case 7:
		doit = (f.psw_SG);
		break;
	}
	if (doit) {
		sys->cpu.pc = a;
	} else {
		sys->cpu.pc += off + 2;
	}
}

void jr_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	sys->cpu.pc += off + 1;

	int cc = ((op1 & 070) >> 3);
	z80_psw_t f = *(get_gp_F(sys));
	int doit = 0;
	switch (cc) {
	case 3:
		doit = 1;
		break;
	case 4:
		doit = (!f.psw_ZR);
		break;
	case 5:
		doit = (f.psw_ZR);
		break;
	case 6:
		doit = (!f.psw_CY);
		break;
	case 7:
		doit = (f.psw_CY);
		break;
	}
	if (doit) {
		sys->cpu.cycles += 5;
		sys->cpu.pc += d;
	}
}

void djnz_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	sys->cpu.cycles += 1;
	sys->cpu.pc += off + 1;

	uint8_t *b = get_gp(sys, gp_B);
	*b -= 1;
	if (*b != 0) {
		sys->cpu.cycles += 5;
		sys->cpu.pc += d;
	}
}

void call_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t a = fetch_imm16(sys, off);
	sys->cpu.pc += off + 2;

	sys->cpu.sp -= 2;
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	*ts = sys->cpu.pc;
	sys->cpu.cycles += 7;
	sys->cpu.pc = a;
}

void rst_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t rv = (op1 & 070);
	sys->cpu.pc += off;

	sys->cpu.sp -= 2;
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	*ts = sys->cpu.pc;
	sys->cpu.cycles += 7;
	sys->cpu.pc = rv;
}

void ccc_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t a = fetch_imm16(sys, off);
	int cc = ((op1 & 070) >> 3);

	z80_psw_t f = *(get_gp_F(sys));
	int doit = 0;
	switch (cc) {
	case 0:
		doit = (!f.psw_ZR);
		break;
	case 1:
		doit = (f.psw_ZR);
		break;
	case 2:
		doit = (!f.psw_CY);
		break;
	case 3:
		doit = (f.psw_CY);
		break;
	case 4:
		doit = (!f.psw_PV);
		break;
	case 5:
		doit = (f.psw_PV);
		break;
	case 6:
		doit = (!f.psw_SG);
		break;
	case 7:
		doit = (f.psw_SG);
		break;
	}
	sys->cpu.pc += off + 2;
	if (doit) {
		sys->cpu.sp -= 2;
		uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
		*ts = sys->cpu.pc;
		sys->cpu.cycles += 7;
		sys->cpu.pc = a;
	}
}

void push_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *qq = get_rp2(sys, (op1 & 0060) >> 4);
	sys->cpu.sp -= 2;
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	*ts = *qq;
	sys->cpu.cycles += 7;
	sys->cpu.pc += off;
}

void pop_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *qq = get_rp2(sys, (op1 & 0060) >> 4);
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	sys->cpu.sp += 2;
	*qq = *ts;
	sys->cpu.cycles += 6;
	sys->cpu.pc += off;
}

void ret_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	sys->cpu.sp += 2;
	sys->cpu.cycles += 6;
	sys->cpu.pc = *ts;
}

void rcc_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	int cc = ((op1 & 070) >> 3);
	z80_psw_t f = *(get_gp_F(sys));
	int doit = 0;
	switch (cc) {
	case 0:
		doit = (!f.psw_ZR);
		break;
	case 1:
		doit = (f.psw_ZR);
		break;
	case 2:
		doit = (!f.psw_CY);
		break;
	case 3:
		doit = (f.psw_CY);
		break;
	case 4:
		doit = (!f.psw_PV);
		break;
	case 5:
		doit = (f.psw_PV);
		break;
	case 6:
		doit = (!f.psw_SG);
		break;
	case 7:
		doit = (f.psw_SG);
		break;
	}
	if (doit) {
		uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
		sys->cpu.sp += 2;
		sys->cpu.cycles += 6;
		sys->cpu.pc = *ts;
	} else {
		sys->cpu.pc += off;
	}
}

void in_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t p = fetch_imm8(sys, off);
	uint8_t *a = get_gp_A(sys);
	io_input(&sys->io, p, a, 1);
	sys->cpu.cycles += 4;
	sys->cpu.pc += off + 1;
}

void out_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t p = fetch_imm8(sys, off);
	uint8_t *a = get_gp_A(sys);
	io_output(&sys->io, p, a, 1);
	sys->cpu.cycles += 4;
	sys->cpu.pc += off + 1;
}

void halt_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	// how to wait for interrupt?
	sys->cpu.pc += off;
}

void rlc_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint16_t v = rlc8a(*a, f);
	*a = v & 0x00ff;
	sys->cpu.pc += off;
}

void rrc_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint16_t v = rrc8a(*a, f);
	*a = v & 0x00ff;
	sys->cpu.pc += off;
}

void ral_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint16_t v = ral8a(*a, f);
	*a = v & 0x00ff;
	sys->cpu.pc += off;
}

void rar_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint16_t v = rar8a(*a, f);
	*a = v & 0x00ff;
	sys->cpu.pc += off;
}

void addi_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t n = fetch_imm8(sys, off);
	uint16_t v = add8(*a, n, f);
	*a = v;
	sys->cpu.pc += off + 1;
}
void adci_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t n = fetch_imm8(sys, off);
	uint16_t v = adc8(*a, n, f);
	*a = v;
	sys->cpu.pc += off + 1;
}
void subi_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t n = fetch_imm8(sys, off);
	uint16_t v = sub8(*a, n, f);
	*a = v;
	sys->cpu.pc += off + 1;
}
void sbbi_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t n = fetch_imm8(sys, off);
	uint16_t v = sbb8(*a, n, f);
	*a = v;
	sys->cpu.pc += off + 1;
}
void andi_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t n = fetch_imm8(sys, off);
	uint16_t v = and8(*a, n, f);
	*a = v;
	sys->cpu.pc += off + 1;
}
void xori_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t n = fetch_imm8(sys, off);
	uint16_t v = xor8(*a, n, f);
	*a = v;
	sys->cpu.pc += off + 1;
}
void ori_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t n = fetch_imm8(sys, off);
	uint16_t v = or8(*a, n, f);
	*a = v;
	sys->cpu.pc += off + 1;
}
void cmpi_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t n = fetch_imm8(sys, off);
	(void)sub8(*a, n, f);
	sys->cpu.pc += off + 1;
}

void daa_op(z80_sys_t *sys, uint8_t op1, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);

	uint8_t av = *a;
	if ((av & 0x0f) > 0x09 || f->psw_HC) {
		av += 0x06;
	}
	if ((av & 0xf0) > 0x90 || f->psw_CY) {
		av += 0x60;
	}

	*a = av; // need to set flags...
	sys->cpu.pc += off;
}
