// $Id: z80_instr2.c,v 1.12 2011-04-06 21:44:04 drmiller Exp $

#include <string.h>

// two-byte opcode instructions
#include "z80_instr2.h"
#include "z80_exec.h"
#include "z80_alu.h"

void bit_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	int r = op2 & 007;
	uint8_t s = 1 << ((op2 & 070) >> 3);
	z80_psw_t *f = get_gp_F(sys);
	if (r == 6) {
		sys->cpu.cycles += 4;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	f->psw_ZR = ((*v & s) == 0);
	sys->cpu.pc += off;
}

void res_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	int r = op2 & 007;
	uint8_t s = 1 << ((op2 & 070) >> 3);
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	*v &= ~s;
	sys->cpu.pc += off;
}

void set_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	int r = op2 & 007;
	uint8_t s = 1 << ((op2 & 070) >> 3);
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	*v |= s;
	sys->cpu.pc += off;
}

void rlc_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int r = op2 & 007;
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	uint16_t rr = rlc8r(*v, f);
	*v = (rr & 0x00ff);
	sys->cpu.pc += off;
}
void rrc_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int r = op2 & 007;
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	uint16_t rr = rrc8r(*v, f);
	*v = (rr & 0x00ff);
	sys->cpu.pc += off;
}
void ral_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int r = op2 & 007;
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	uint16_t rr = ral8r(*v, f);
	*v = (rr & 0x00ff);
	sys->cpu.pc += off;
}
void rar_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int r = op2 & 007;
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	uint16_t rr = rar8r(*v, f);
	*v = (rr & 0x00ff);
	sys->cpu.pc += off;
}
void sla_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int r = op2 & 007;
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	uint16_t rr = sla8r(*v, f);
	*v = (rr & 0x00ff);
	sys->cpu.pc += off;
}
void sra_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int r = op2 & 007;
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	uint16_t rr = sra8r(*v, f);
	*v = (rr & 0x00ff);
	sys->cpu.pc += off;
}
void srl_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int r = op2 & 007;
	if (r == 6) {
		sys->cpu.cycles += 7;
		v = mem_get_addr(&sys->mem, *(get_rp(sys, rp_HL)));
	} else {
		v = get_gp(sys, r);
	}
	uint16_t rr = srl8r(*v, f);
	*v = (rr & 0x00ff);
	sys->cpu.pc += off;
}
void inc_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	sys->cpu.cycles += 7;
	int cy = f->psw_CY;
	uint16_t rr = add8(*v, 1, f);
	f->psw_CY = cy;
	*v = (rr & 0x00ff);
	sys->cpu.pc += off + 1;
}
void dec_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	sys->cpu.cycles += 7;
	int cy = f->psw_CY;
	uint16_t rr = sub8(*v, 1, f);
	f->psw_CY = cy;
	*v = (rr & 0x00ff);
	sys->cpu.pc += off + 1;
}
void inx_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	sys->cpu.cycles += 2;
	uint16_t *v = &sys->cpu.ix[x];
	++(*v);
	sys->cpu.pc += off;
}
void dcx_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	sys->cpu.cycles += 2;
	uint16_t *v = &sys->cpu.ix[x];
	--(*v);
	sys->cpu.pc += off;
}
void dad_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	int pp = (op2 & 0060) >> 4;
	z80_psw_t *f = get_gp_F(sys);
	sys->cpu.cycles += 7;
	uint16_t *v = &sys->cpu.ix[x];
	uint16_t qq = *(get_rp3(sys, pp, x));
	uint16_t rr = add16(*v, qq, f);
	*v = rr;
	sys->cpu.pc += off;
}
void mvi_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *v;
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t n = fetch_imm8(sys, off + 1);
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	*v = n;
	sys->cpu.cycles += 5;
	sys->cpu.pc += off + 2;
}
void lxi_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t nn = fetch_imm16(sys, off);
	uint16_t *v = &sys->cpu.ix[x];
	*v = nn;
	sys->cpu.pc += off + 2;
}
void lixd_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t nn = fetch_imm16(sys, off);
	uint16_t *v = &sys->cpu.ix[x];
	*v = mem_get_word(&sys->mem, nn);
	sys->cpu.cycles += 6;
	sys->cpu.pc += off + 2;
}
void sixd_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t nn = fetch_imm16(sys, off);
	uint16_t *v = &sys->cpu.ix[x];
	*(mem_get_addr(&sys->mem, nn)) = *v;
	sys->cpu.cycles += 6;
	sys->cpu.pc += off + 2;
}
void mov_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	int r = (op2 & 0007);
	if (r == 6) {
		r = (op2 & 0070) >> 3;
		uint8_t *b = get_gp(sys, r);
		*b = *v;
	} else {
		uint8_t *b = get_gp(sys, r);
		*v = *b;
	}
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void add_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint8_t *a = get_gp_A(sys);
	uint16_t r = add8(*a, *v, f);
	*a = r & 0x00ff;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void adc_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint8_t *a = get_gp_A(sys);
	uint16_t r = adc8(*a, *v, f);
	*a = r & 0x00ff;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void sub_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint8_t *a = get_gp_A(sys);
	uint16_t r = sub8(*a, *v, f);
	*a = r & 0x00ff;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void sbb_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint8_t *a = get_gp_A(sys);
	uint16_t r = sbb8(*a, *v, f);
	*a = r & 0x00ff;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void and_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint8_t *a = get_gp_A(sys);
	uint16_t r = and8(*a, *v, f);
	*a = r & 0x00ff;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void xor_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint8_t *a = get_gp_A(sys);
	uint16_t r = xor8(*a, *v, f);
	*a = r & 0x00ff;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void or_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint8_t *a = get_gp_A(sys);
	uint16_t r = or8(*a, *v, f);
	*a = r & 0x00ff;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void cmp_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	z80_psw_t *f = get_gp_F(sys);
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t d = (int8_t)fetch_imm8(sys, off);
	uint8_t *v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint8_t *a = get_gp_A(sys);
	(void)sub8(*a, *v, f);
	sys->cpu.cycles += 8;
	sys->cpu.pc += off + 1;
}
void pop_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t *qq = &sys->cpu.ix[x];
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	sys->cpu.sp -= 2;
	*qq = *ts;
	sys->cpu.cycles += 6;
	sys->cpu.pc += off;
}
void push_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t *qq = &sys->cpu.ix[x];
	sys->cpu.sp -= 2;
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	*ts = *qq;
	sys->cpu.cycles += 7;
	sys->cpu.pc += off;
}
void spix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t *ix = &sys->cpu.ix[x];
	sys->cpu.sp = *ix;
	sys->cpu.cycles += 2;
	sys->cpu.pc += off;
}
void xtix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t *ix = &sys->cpu.ix[x];
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	uint16_t t = *ix;
	*ix = *ts;
	*ts = t;
	sys->cpu.cycles += 15;
	sys->cpu.pc += off;
}
void pcix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	int x = (op1 & 0040) >> 5;      // I_IX or I_IY
	uint16_t *ix = &sys->cpu.ix[x];
	sys->cpu.pc = *ix;
}

void in_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t rn = (op2 & 0070) >> 3;
	uint8_t *r = get_gp(sys, rn);
	uint8_t p = *(get_gp(sys, gp_C));
	z80_psw_t *f = get_gp_F(sys);
	uint8_t t;
	io_input(&sys->io, p, &t, 1);
	(void)or8(t, 0, f);
	if (rn != 6) {
		*r = t;
	}
	sys->cpu.cycles += 4;
	sys->cpu.pc += off;
}
void out_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *r = get_gp(sys, (op2 & 0070) >> 3);
	uint8_t p = *(get_gp(sys, gp_C));
	io_output(&sys->io, p, r, 1);
	sys->cpu.cycles += 4;
	sys->cpu.pc += off;
}
void reti_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	sys->cpu.sp += 2;
	// todo: interrupt flags...
	sys->cpu.cycles += 6;
	sys->cpu.pc = *ts;
	sys->cpu.pc += off;
}
void retn_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint16_t *ts = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	sys->cpu.sp += 2;
	// todo: interrupt flags...
	sys->cpu.cycles += 6;
	sys->cpu.pc = *ts;
	sys->cpu.pc += off;
}
void dadc_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint16_t *r = get_rp(sys, (op1 & 0060) >> 4);
	uint16_t *h = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);

	*h = adc16(*h, *r, f);
	sys->cpu.cycles += 7;
	sys->cpu.pc += off;
}
void dsbc_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint16_t *r = get_rp(sys, (op1 & 0060) >> 4);
	uint16_t *h = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);

	*h = sbc16(*h, *r, f);
	sys->cpu.cycles += 7;
	sys->cpu.pc += off;
}
void neg_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	z80_psw_t *f = get_gp_F(sys);
	uint16_t v = sub8(0, *a, f);
	*a = v;
	sys->cpu.pc += off;
}
void im0_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	sys->cpu.im = 0;
	sys->cpu.pc += off;
}
void im1_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	sys->cpu.im = 1;
	sys->cpu.pc += off;
}
void im2_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	sys->cpu.im = 2;
	sys->cpu.pc += off;
}
void rrd_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint16_t *h = get_rp(sys, rp_HL);
	uint8_t *m = mem_get_addr(&sys->mem, *h);
	z80_psw_t *f = get_gp_F(sys);
	uint32_t v = (*a << 8) | *m;
	v = rrd8r(v, f);
	*a = (v >> 8);
	*m = (v & 0x00ff);
	sys->cpu.cycles += 10;
	sys->cpu.pc += off;
}
void rld_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint16_t *h = get_rp(sys, rp_HL);
	uint8_t *m = mem_get_addr(&sys->mem, *h);
	z80_psw_t *f = get_gp_F(sys);
	uint32_t v = (*a << 8) | *m;
	v = rld8r(v, f);
	*a = (v >> 8);
	*m = (v & 0x00ff);
	sys->cpu.cycles += 10;
	sys->cpu.pc += off;
}
void mov_i_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	if (op2 & 0020) {
		*a = sys->cpu.iv;
	} else {
		sys->cpu.iv = *a;
	}
	sys->cpu.cycles += 1;
	sys->cpu.pc += off;
}
void mov_r_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	if (op2 & 0020) {
		*a = sys->cpu.re;
	} else {
		sys->cpu.re = *a;
	}
	sys->cpu.cycles += 1;
	sys->cpu.pc += off;
}
void sxxd_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint16_t nn = fetch_imm16(sys, off);
	uint16_t *dd = get_rp(sys, (op2 & 0060) >> 4);
	uint16_t *mm = (uint16_t *)mem_get_addr(&sys->mem, nn);
	*mm = *dd;
	sys->cpu.cycles += 6;
	sys->cpu.pc += off + 2;
}
void lxxd_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint16_t nn = fetch_imm16(sys, off);
	uint16_t *dd = get_rp(sys, (op2 & 0060) >> 4);
	uint16_t *mm = (uint16_t *)mem_get_addr(&sys->mem, nn);
	*dd = *mm;
	sys->cpu.cycles += 6;
	sys->cpu.pc += off + 2;
}
void ldi_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint16_t *bc = get_rp(sys, rp_BC);
	uint16_t *de = get_rp(sys, rp_DE);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mde = mem_get_addr(&sys->mem, *de);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	*mde = *mhl;
	--(*bc);
	++(*hl);
	++(*de);
	f->psw_PV = (*bc != 0);
	f->psw_NN = 0;
	f->psw_HC = 0;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off;
}
void ldd_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint16_t *bc = get_rp(sys, rp_BC);
	uint16_t *de = get_rp(sys, rp_DE);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mde = mem_get_addr(&sys->mem, *de);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	*mde = *mhl;
	--(*bc);
	--(*hl);
	--(*de);
	f->psw_PV = (*bc != 0);
	f->psw_NN = 0;
	f->psw_HC = 0;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off;
}
void ldir_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	// does not handle wrapping around
	uint16_t *bc = get_rp(sys, rp_BC);
	uint16_t *de = get_rp(sys, rp_DE);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mde = mem_get_addr(&sys->mem, *de);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	memcpy(mde, mhl, *bc);
	*hl += *bc;
	*de += *bc;
	sys->cpu.cycles += 13 * *bc - 5;
	*bc = 0;
	f->psw_PV = 0;
	f->psw_NN = 0;
	f->psw_HC = 0;
	sys->cpu.pc += off;
}
void lddr_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	// does not handle wrapping around
	uint16_t *bc = get_rp(sys, rp_BC);
	uint16_t *de = get_rp(sys, rp_DE);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mde = mem_get_addr(&sys->mem, *de);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	memcpy(mde, mhl, *bc);
	*hl -= *bc;
	*de -= *bc;
	sys->cpu.cycles += 13 * *bc - 5;
	*bc = 0;
	f->psw_PV = 0;
	f->psw_NN = 0;
	f->psw_HC = 0;
	sys->cpu.pc += off;
}
void cpi_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint16_t *bc = get_rp(sys, rp_BC);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	(void)sub8(*a, *mhl, f);
	--(*bc);
	++(*hl);
	f->psw_PV = (*bc != 0);
	sys->cpu.cycles += 8;
	sys->cpu.pc += off;
}
void cpd_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t *a = get_gp_A(sys);
	uint16_t *bc = get_rp(sys, rp_BC);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	(void)sub8(*a, *mhl, f);
	--(*bc);
	--(*hl);
	f->psw_PV = (*bc != 0);
	sys->cpu.cycles += 8;
	sys->cpu.pc += off;
}
void cpir_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	// does not handle wrapping around
	uint8_t *a = get_gp_A(sys);
	uint16_t *bc = get_rp(sys, rp_BC);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	uint8_t *st = mhl;
	uint8_t aa = *a;
	uint16_t bb = *bc;
	uint8_t m;
	do {
		--bb;
		m = *st++;
	} while (bb && aa != m);
	(void)sub8(aa, m, f);
	bb = *bc - bb;
	*bc -= bb;
	*hl += bb;
	f->psw_PV = (*bc != 0);
	sys->cpu.cycles += 13 * bb - 5;
	sys->cpu.pc += off;
}
void cpdr_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	// does not handle wrapping around
	uint8_t *a = get_gp_A(sys);
	uint16_t *bc = get_rp(sys, rp_BC);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	uint8_t *st = mhl;
	uint8_t aa = *a;
	uint16_t bb = *bc;
	uint8_t m;
	do {
		--bb;
		m = *st--;
	} while (bb && aa != m);
	(void)sub8(aa, m, f);
	bb = *bc - bb;
	*bc -= bb;
	*hl -= bb;
	f->psw_PV = (*bc != 0);
	sys->cpu.cycles += 13 * bb - 5;
	sys->cpu.pc += off;
}
void ini_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t p = *(get_gp(sys, gp_C));
	uint8_t *b = get_gp(sys, gp_B);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	io_input(&sys->io, p, mhl, 1);
	--(*b);
	++(*hl);
	f->psw_ZR = (*b == 0);
	f->psw_NN = 1;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off;
}
void ind_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t p = *(get_gp(sys, gp_C));
	uint8_t *b = get_gp(sys, gp_B);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	io_input(&sys->io, p, mhl, 1);
	--(*b);
	--(*hl);
	f->psw_ZR = (*b == 0);
	f->psw_NN = 1;
	sys->cpu.cycles += 8;
	sys->cpu.pc += off;
}
void inir_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	// does not handle wrapping around
	uint8_t p = *(get_gp(sys, gp_C));
	uint8_t *b = get_gp(sys, gp_B);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	io_input(&sys->io, p, mhl, *b);
	sys->cpu.cycles += 13 * *b - 5;
	*hl += *b;
	*b = 0;
	f->psw_ZR = 1;
	f->psw_NN = 1;
	sys->cpu.pc += off;
}
void indr_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	// does not handle wrapping around
	uint8_t p = *(get_gp(sys, gp_C));
	uint8_t *b = get_gp(sys, gp_B);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	io_input(&sys->io, p, mhl, *b);
	sys->cpu.cycles += 13 * *b - 5;
	*hl -= *b;
	*b = 0;
	f->psw_ZR = 1;
	f->psw_NN = 1;
	sys->cpu.pc += off;
}
void outi_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t p = *(get_gp(sys, gp_C));
	uint8_t *b = get_gp(sys, gp_B);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	io_output(&sys->io, p, mhl, 1);
	sys->cpu.cycles += 8;
	--(*b);
	++(*hl);
	f->psw_ZR = (*b == 0);
	f->psw_NN = 1;
	sys->cpu.pc += off;
}
void outd_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	uint8_t p = *(get_gp(sys, gp_C));
	uint8_t *b = get_gp(sys, gp_B);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	io_output(&sys->io, p, mhl, 1);
	sys->cpu.cycles += 8;
	--(*b);
	--(*hl);
	f->psw_ZR = (*b == 0);
	f->psw_NN = 1;
	sys->cpu.pc += off;
}
void outir_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	// does not handle wrapping around
	uint8_t p = *(get_gp(sys, gp_C));
	uint8_t *b = get_gp(sys, gp_B);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	io_output(&sys->io, p, mhl, *b);
	sys->cpu.cycles += 13 * *b - 5;
	*hl += *b;
	*b = 0;
	f->psw_ZR = 1;
	f->psw_NN = 1;
	sys->cpu.pc += off;
}
void outdr_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint16_t off) {
	// does not handle wrapping around
	uint8_t p = *(get_gp(sys, gp_C));
	uint8_t *b = get_gp(sys, gp_B);
	uint16_t *hl = get_rp(sys, rp_HL);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *mhl = mem_get_addr(&sys->mem, *hl);
	io_output(&sys->io, p, mhl, *b);
	sys->cpu.cycles += 13 * *b - 5;
	*hl -= *b;
	*b = 0;
	f->psw_ZR = 1;
	f->psw_NN = 1;
	sys->cpu.pc += off;
}
