// $Id: z80_instr3.c,v 1.8 2011-04-03 17:23:52 drmiller Exp $

// three-byte opcode instructions
#include "z80_instr3.h"
#include "z80_exec.h"
#include "z80_alu.h"

void bit_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	z80_psw_t *f = get_gp_F(sys);
	uint8_t *v;
	sys->cpu.cycles += 5;
	// assert((op3 & 007) == R_M);
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	uint8_t s = 1 << ((op2 & 070) >> 3);
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	f->psw_ZR = ((*v & s) == 0);
	sys->cpu.pc += off;
}

void res_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	uint8_t s = 1 << ((op2 & 070) >> 3);
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	*v &= ~s;
	sys->cpu.pc += off;
}

void set_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	uint8_t s = 1 << ((op2 & 070) >> 3);
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	*v |= s;
	sys->cpu.pc += off;
}

void rlc_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint16_t r = rlc8r(*v, f);
	*v = (r & 0x00ff);
	sys->cpu.pc += off;
}

void rrc_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint16_t r = rrc8r(*v, f);
	*v = (r & 0x00ff);
	sys->cpu.pc += off;
}
void ral_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint16_t r = ral8r(*v, f);
	*v = (r & 0x00ff);
	sys->cpu.pc += off;
}
void rar_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint16_t r = rar8r(*v, f);
	*v = (r & 0x00ff);
	sys->cpu.pc += off;
}
void sla_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint16_t r = sla8r(*v, f);
	*v = (r & 0x00ff);
	sys->cpu.pc += off;
}
void sra_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint16_t r = sra8r(*v, f);
	*v = (r & 0x00ff);
	sys->cpu.pc += off;
}
void srl_ix_op(z80_sys_t *sys, uint8_t op1, uint8_t op2, uint8_t op3, uint16_t off) {
	uint8_t *v;
	z80_psw_t *f = get_gp_F(sys);
	uint16_t d = (int8_t)fetch_imm8(sys, off - 2);
	sys->cpu.cycles += 8;
	int x = (op1 & 0040) >> 5;	// I_IX or I_IY
	v = mem_get_addr(&sys->mem, sys->cpu.ix[x] + d);
	uint16_t r = srl8r(*v, f);
	*v = (r & 0x00ff);
	sys->cpu.pc += off;
}
