// $Id: z80_sim_cmd.c,v 1.2 2011-04-08 23:11:31 drmiller Exp $

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "z80_sys.h"
#include "z80_exec.h"

extern int diz80(char *buf, uint8_t *t);

static int _dump(z80_sys_t *sys, char *line) {
	char buf[1024];

	uint8_t *pc = mem_get_addr(&sys->mem, sys->cpu.pc);
	diz80(buf, pc);
	fprintf(stderr, "PC = %04x [ %s ]\n", sys->cpu.pc, buf);
	uint16_t *sp = (uint16_t *)mem_get_addr(&sys->mem, sys->cpu.sp);
	fprintf(stderr, "SP = %04x [ %04x %04x %04x %04x ]\n",
			sys->cpu.sp, sp[0], sp[1], sp[2], sp[3]);

	uint8_t a = sys->cpu.psw.af[sys->cpu.psw.af_f].af[af_A];
	z80_psw_t *f = (z80_psw_t *)&sys->cpu.psw.af[sys->cpu.psw.af_f].af[af_F];
	uint16_t bc = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f].xx[rp_BC];
	uint16_t de = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f].xx[rp_DE];
	uint16_t hl = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f].xx[rp_HL];
	fprintf(stderr, "A  = %02x [ %s ]\n", a, get_psw_str(sys, f));
	fprintf(stderr, "BC = %04x DE = %04x HL = %04x\n", bc, de, hl);

	a = sys->cpu.psw.af[sys->cpu.psw.af_f ^ 1].af[af_A];
	f = (z80_psw_t *)&sys->cpu.psw.af[sys->cpu.psw.af_f ^ 1].af[af_F];
	bc = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f ^ 1].xx[rp_BC];
	de = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f ^ 1].xx[rp_DE];
	hl = sys->cpu.gp_regs.gp[sys->cpu.gp_regs.gp_f ^ 1].xx[rp_HL];
	fprintf(stderr, "A' = %02x [ %s ]'\n", a, get_psw_str(sys, f));
	fprintf(stderr, "BC'= %04x DE'= %04x HL'= %04x\n", bc, de, hl);

	fprintf(stderr, "IX = %04x IY = %04x I = %02x R = %02x\n",
				sys->cpu.ix[ix_X], sys->cpu.ix[ix_Y],
				sys->cpu.iv, sys->cpu.re);
	return 0;
}

static int _trace(z80_sys_t *sys, char *line) {
	char *s;
	s = strtok(NULL, " \t");
	if (s) {
		if (strcmp(s, "on") == 0) {sys->trace = 1;}
		else if (strcmp(s, "off") == 0) {sys->trace = 0;}
		else {
			FILE *fp = fopen(s, "a");
			if (!fp) {
				perror(s);
				return 0;
			}
			sys->trace = 1;
			sys->trc_fp = fp;
		}
	} else {
		sys->trace = !sys->trace;
	}
	if (!sys->trace && sys->trc_fp != stderr) {
		fclose(sys->trc_fp);
		sys->trc_fp = stderr;
	}
	if (sys->trace) {
		printf("trace is on (%s)\n",
			sys->trc_fp == stderr ? "stderr" : "file");
	} else {
		printf("trace is off\n");
	}
	return 0;
}

static int _go(z80_sys_t *sys, char *line) {
	sys->run = 1;
	printf("resuming at %04x\n", sys->cpu.pc);
	return 0;
}

static int _help(z80_sys_t *sys, char *line) {
	printf(	"Z80-SIM Commands:\n"
		"\tquit\tEnd simulation\n"
		"\ttrace\tToggle trace on/off\n"
		"\tdump\tDump processor state/registers\n"
		"\tgo\tResume program at current PC\n"
		"\thelp\tDisplay this help\n"
		);
	return 0;
}

struct {
	char *cmd;
	int (*func)(z80_sys_t *sys, char *line);
} commands[] = {
	{ "quit", NULL },
	{ "trace", _trace },
	{ "dump", _dump },
	{ "go", _go },
	{ "help", _help },
};
#define NUM_CMDS	(sizeof(commands) / sizeof(commands[0]))

void sys_command(z80_sys_t *sys) {
	char buf[128];
	int x;

	printf("%% ");
	fflush(stdout);
	char *s = fgets(buf, sizeof(buf), stdin);
	if (!s) {
		printf("z80-sim done.\n");
		exit(0);
	}
	x = strlen(s);
	if (!x) return;
	--x;
	if (s[x] == '\n') s[x] = '\0';
	s = strtok(buf, " \t");
	for (x = 0; x < NUM_CMDS; ++x) {
		if (strcmp(s, commands[x].cmd) == 0) {
			break;
		}
	}
	if (!(x < NUM_CMDS)) {
		printf("%s ?\n", s);
		return;
	}
	if (!commands[x].func) {
		printf("z80-sim done.\n");
		exit(0);
	}
	int rc = commands[x].func(sys, buf);
	if (rc) {
		// what is this for?
		exit(1);
	}
}
